/*
  This file is part of a patch to SDL_mixer 1.2 to support MP3 playback
  using libmpg123. It is part of the Pygame project (www.pygame.org).

  pygame - Python Game Library
  Copyright (C) 2000-2001  Pete Shinners

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Lenard Lindstrom
  len-l@telus.net


  SDL_mixer:  An audio mixer library based on the SDL library
  Copyright (C) 1997-2012 Sam Lantinga <slouken@libsdl.org>


  The libmpg123 interface for music.c. See music_mp3_mpg123.h.
*/

/* $Id$ */

#ifdef MP3_MPG123_MUSIC

/* Support for MP3 (MPEG-1 or MPEG-2 layer III) music streams */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SDL_mixer.h"
#include <mpg123.h>
#include "music_mp3_mpg123.h"

#define READ_BLOCK_SIZE 4096

/* General stream state. */
typedef enum {
    /* There are samples in the libmpg123 stream. */
    STATE_ACTIVE,

    /* The libmpg123 stream is finished. This allows the music stream to
       close on the read following the one returning the last bytes in
       the sample buffer. */
    STATE_DONE,

    /* A function call failure has left the stream in an irrecoverable
       state. */
    STATE_ERROR
} state;

/* A music stream */
typedef struct stream_mpg123_s {
    /* MP3 encoded raw source. */
    SDL_RWops *rw;

    /* Boolean: whether or not to close *rw on completion.*/
    int freerw;

    /* Boolean: whether or not the music stream is playing, ie,
       will return samples when read. */
    int playing;

    /* General stream state. */
    state playback;

    /* SDL_mixer volume between 0 and MIX_MAX_VOLUME inclusive. 
       Volume adjustment is done when samples are copied from the music
       streams samples buffer into the external read buffer. */
    int volume;

    /* Reads bytes from *rw, returning an uncompressed stream of samples. */
    mpg123_handle *decoder;

    /* The number of bytes to retrieve with each read of *decoder. */
    size_t read_block_size;

    /* The in-place conversion that transforms the samples returned
       by *decoder into the final specification of the music stream.
       ctv.buf points to the allocated sample buffer, which collects
       samples from *decoder and from which converted samples are
       read. Samples are buffered at full volume. */
    SDL_AudioCVT cvt;

    /* The total amount of memory allocated to the samples buffer.
       This is at least enough to hold read_block_size samples after
       convertion. The size will be increased as needed. It will not
       be reduced. */
    size_t buffer_alloc_size;

    /* The location of the next unread sample in the samples buffer. */
    Uint8 *buffer_start;

    /* The amount of of buffered samples, in bytes, remaining to be
       read, from 0 to buffer_alloc_size inclusive. */
    size_t buffer_len;
} stream_mpg123;

/* The music stream description. These are the relevant fields of
   the playing streams SDL_AudioSpec description. */
typedef struct {
    /* SDL_AudioSpec format, freq, and channels fields. */
    Uint16 mix_format;
    int mix_freq;
    Uint8 mix_channels;

    /* The above fields in libmpg123 terms. This is what the MP3 decoder
       returns. Further convertion into the required SDL specification
       may be necessary. */
    int mpg123_encoding;
    long mpg123_rate;
    int mpg123_channels;
} stream_format;

static int add_samples(stream_mpg123 *s);
static int move_samples(stream_mpg123 *s, Uint8 *buffer, int len);
static size_t get_buffered_len(stream_mpg123 *s);
static void clear_buffer(stream_mpg123 *s);
static ssize_t read_callback(void *stream, void *buffer, size_t len);
static off_t seek_callback(void *stream, off_t offset, int whence);
static stream_mpg123 *new_stream(void);
static int add_decoder(stream_mpg123 *s);
static int add_handle_mpg123(stream_mpg123 *s);
static int SDL_encoding_as_mpg123(int format);
static int set_convertion(stream_mpg123 *s);
static int adjust_buffer(stream_mpg123 *s);
static void set_mpg123_error(stream_mpg123 *s, const char *msg);

/* Required output format. Set by init_mpg123(). The format will remain
   unchanged until the mixer quits. It is assumed init_mpg123() will
   be called again with the new format if the mixer is restarted.
   */
static stream_format target = {0, 0, 0, 0, 0, 0};

/* The mpg123 library does not track if mpg123_init() was called before.
   Do it here. */
static int was_init_libmpg123 = 0;

/* Stream States

   In general, the state of an 'stream_mpg123' instance between API function
   calls is defined by the fields 'playable', 'playing', and 'buffer_size'.
   As a simplification, 'buffer_size' can be approximated as a boolean value,
   true when non-zero, false when 0. Now an open music_mp3_mpg123 stream state
   can be represented as {s, p, b}, where s is one of the 'state' enumeration
   values, p is a boolean (0 or 1) which is true for a playing stream, and b is
   a boolean which is false for an empty samples buffer.

   A music_mp3_mpg123 public API function call is a state change of the stream. 
   In the case an outside factor can decide final state of a stream, the
   function call can be qualified. The read_mpg123() and position_mpg123()
   functions can fail, which may be qualified as an error (ERR). Function
   read_mpg123() will also notes when the last of the libmpg123 'decoder'
   stream has been read, qualified as end-of-file (EOF). So the action of
   a function call on a stream can be properly represented with the addition
   of ERR or EOF when the final state is not determined entirely by the
   initial state.

   To complete the picture the stream may be uninitialized, open_mgp123() has
   yet to be called, or closed, by calling close_mpg123(). In particular,
   closing a stream invalidates any pointers to it. So two new states,
   {Begin} and {End}, are added to represent the lifetime of the stream in
   relation to the lifetime of a 'stream_mpg123' pointer.


   State Table for an music_mp3_mpg123 MP3 Stream:
   (organized by API function)

   Action                   From State              To State
   ---------------------------------------------------------------------------
                          |                       |
   open_rw_mpg123()       | Begin                 | {STATE_ACTIVE, 0, 0}
   open_rw_mpg123()       | Begin                 | End
                          |                       |
   close_rw_pmg123()      | {_, _, _}             | End
                          |                       |
   play_mpg123()          | {STATE_ACTIVE, 0, b}  | {STATE_ACTIVE, 1, b}
                          | {STATE_DONE, 0, 1}    | {STATE_DONE, 1, 1}
                          | {s, p, b}             | {s, p, b}
                          |                       |
   stop_mpg123()          | {s, _, b}             | {s, 0, b}
                          |                       |
   read_mpg123()          | {STATE_ACTIVE, 1, _}  | {STATE_ACTIVE, 1, b}  (1)
                          | {s, p, b}             | {s, p, b}
   read_mpg123() EOF      | {STATE_ACTIVE, 1, 0}  | {STATE_DONE, s, b}    (1)(2)
                          | {STATE_DONE, 1, 1}    | {STATE_DONE, 1, b}    (1)
                          | {STATE_DONE, 1, 0}    | {STATE_DONE, 0, 0}
                          | {s, p, b}             | {s, p, b}
   read_mpg123() ERR      | {_, _, _}             | {STATE_ERROR, 0, 0}
                          |                       |
   position_mpg123()      | {STATE_ACTIVE, p, _}  | {STATE_ACTIVE, p, 0}
                          | {STATE_DONE, p, _}    | {STATE_ACTIVE, p, 0}
                          | {s, p, b}             | {s, p, b}
   position_mpg123() ERR  | {_, _, _}             | {STATE_ERROR, 0, 0}
                          |                       |
   ---------------------------------------------------------------------------

   Action: A qualified function call.
   From: The stream state passed to the function. When an action has
         multiple 'From' choices, each pattern takes precedence over the
         ones below it. Pattern literals, such as STATE_ACTIVE or 1, represent
         an exact match. Variables s, p, b, and _ match any valid value for its
         field. 
   To: The stream states returned by the function.
       
   (1) Because the b state is an approximation of 'buffer_size', it depends on
       hidden state. This makes its final value indeterminate with regards to
       this action on the initial state. Fortunately, the field is not relevant
       to the overall stream state at this time. However, b does indicate when
       the samples buffer is cleared. It also stops the stream when it has
       reached the end. In both cases b has a well defined value of 0.
   (2) Either both p and b are 1 or both are 0. Normally a stream switches to
       STATE_DONE when the last samples are returned by mpg123_read(), as it
       signals end-of-stream at this time. So there are buffered samples to
       be played, even though the MP3 stream has finished. However, if the
       MP3 stream is positioned past the end, then mpg123_read() does signal
       end-of-stream while returning no samples.

   Forbidden states:
      {STATE_ERROR, 1, _}
      {STATE_ERROR, 0, 1}
      {STATE_ACTIVE, 0, _}
      {STATE_DONE, 0, 1}
   */

/******************  The Public API  ********************/

int
init_mpg123(SDL_AudioSpec *output)
{
    target.mix_format = output->format;
    target.mpg123_encoding = SDL_encoding_as_mpg123(target.mix_format);
    target.mix_freq = output->freq;
    target.mix_channels = output->channels;
    target.mpg123_channels = (target.mix_channels == 1) ?
                                  MPG123_MONO : MPG123_STEREO;
    return (target.mpg123_encoding == 0) ? -1 : 0;
}

void
quit_mpg123(void)
{
    if (was_init_libmpg123) {
        mpg123_exit();
        was_init_libmpg123 = 0;
    }
}

void
setvolume_mpg123(stream_mpg123 *music, int volume)
{
    music->volume = volume;
}

stream_mpg123 *
open_RW_mpg123(SDL_RWops *rw, int freerw)
{
    stream_mpg123 *s;

    /* Ensure that init_mpg123() was called, and the output format
       is valid. */
    if (target.mpg123_encoding == 0) {
        if (SDL_GetError()[0] == '\0') {
            /* No error set, so init_mpg123() was not called. */
            Mix_SetError("init_mpg123() was not called first.");
        }
        return NULL;
    }

    /* Establish the output format. This remains constant. */
    s = new_stream();
    if (!s) {
        if (freerw) {
            SDL_RWclose(rw);
        }
        return NULL;
    }

    /* Make close_mpg123() responsible for rw. */
    s->rw = rw;
    s->freerw = freerw;

    /* Get the source stream sample rate. This may change during playback. */
    if (add_decoder(s)) {
        close_mpg123(s);
        return NULL;
    }

    /* Allocate the samples buffer according to initial source sample rate.
       Buffer size may increase if the sample rate changes. */
    s->read_block_size = READ_BLOCK_SIZE; 
    if (set_convertion(s)) {
        close_mpg123(s);
        return NULL;
    }

    s->playback = STATE_ACTIVE;
    return s;
}

void
close_mpg123(stream_mpg123 *s)
{
    if (s->decoder) {
        mpg123_close(s->decoder);
        mpg123_delete(s->decoder);
    }
    if (s->rw && s->freerw) {
        SDL_RWclose(s->rw);
    }
    if (s->cvt.buf) {
        SDL_free(s->cvt.buf);
    }
    SDL_free(s);
}

void
play_mpg123(stream_mpg123 *s)
{
    if (s->playback == STATE_ACTIVE ||
        (s->playback == STATE_DONE && get_buffered_len(s) > 0)) {
        s->playing = 1;
    }
}

int
playing_mpg123(stream_mpg123 *s)
{
    return s->playing;
}

void
stop_mpg123(stream_mpg123 *s)
{
    s->playing = 0;
}

int
read_mpg123(stream_mpg123 *s, Uint8 *buffer, int len)
{
    size_t size;
    int len_moved;
    int reading = (len > 0 && playing_mpg123(s));

    while (reading) {
        size = get_buffered_len(s);
        if (size == 0) {
            size = add_samples(s);
        }

        if (size > 0) {
            len_moved = move_samples(s, buffer, len);
            buffer += len_moved;
            len -= len_moved;
            reading = (len != 0 && s->playback == STATE_ACTIVE);
        }
        else {
            /* Stop looping if no samples are available. */
            reading = 0;
        }
    }

    return len;
}

void
position_mpg123(stream_mpg123 *s, double time)
{
    off_t frame = 0;

    if (s->playback == STATE_ERROR) {
        /* Nothing to do. */
        return;
    }

    if (time > 0.0) {
        /* Advance the stream by time seconds. */
        frame = mpg123_timeframe(s->decoder, time);
        if (frame >= 0) {
            frame = mpg123_seek_frame(s->decoder, frame, SEEK_CUR);
        }
    }
    else if (time == 0.0) {
        /* Rewind the stream. */
        frame = mpg123_seek_frame(s->decoder, 0, SEEK_SET);
    }

    if (frame < 0) {
        /* An mpg123 error occurred. */
        set_mpg123_error(s, "");
        return;
    }
        
    /* Do not play any buffered samples. */
    clear_buffer(s);

    /* Ensure a stopped stream can be started. */
    if (!s->playing) {
        /* Any completed stream is reset. */
        s->playback = STATE_ACTIVE;
    }
}

/******************  Support Functions  *****************/

/* A libmpg123 low-level I/O read function. */
static ssize_t
read_callback(void *stream, void *buffer, size_t len)
{
    return SDL_RWread((SDL_RWops *)stream, buffer, 1, (int)len);
}

/* A libmpg123 low-lever I/O seek function. */
static off_t
seek_callback(void *stream, off_t offset, int whence)
{
    return SDL_RWseek((SDL_RWops *)stream, (int)offset, whence);
}

/* Return a new stream in the error state, or NULL on failure. It is in an
   internal transitional state; there is no RWops stream, libmpg123 decoder
   handle, or SDL convert buffer. It would be an "error" if the new instance
   escapes to the outside before becoming active. */
static stream_mpg123 *
new_stream(void)
{
    stream_mpg123 *s = (stream_mpg123 *)SDL_malloc(sizeof(stream_mpg123));

    if (!s) {
        SDL_OutOfMemory();
        return NULL;
    }
    s->rw = NULL;
    s->freerw = 0;
    s->playing = 0;
    s->playback = STATE_ERROR;
    s->volume = MIX_MAX_VOLUME;
    s->decoder = NULL;
    s->read_block_size = 0;
    s->cvt.buf = NULL;
    s->buffer_alloc_size = 0;
    s->buffer_start = NULL;
    s->buffer_len = 0;
    return s;
}

/* Open a new libmpg123 handle to decode the MP3 stream returned by the
   RWops file. Return 0 on success, -1 on failure. The libmpg123 handle
   is configured to return samples of the correct size and sign. The number
   of channels will be either 1 or 2. */
static int
add_decoder(stream_mpg123 *s)
{
    int err;

    /* Since mpg123 does not track if it was initialized, do it here. */
    if (!was_init_libmpg123) {
        err = mpg123_init();
        if (err != MPG123_OK) {
            Mix_SetError(mpg123_plain_strerror(err));
            s->playback = STATE_ERROR;
            return -1;
        }
        was_init_libmpg123 = 1;
    }

    if (add_handle_mpg123(s)) {
        return -1;
    }

    /* Check that the MP3 stream and out format are acceptable. */
    if (mpg123_open_handle(s->decoder, s->rw) != MPG123_OK) {
        set_mpg123_error(s, "");
        mpg123_delete(s->decoder);
        s->decoder = NULL;
        return -1;
    }
    return 0;
}

/* Return new smp123 decoder handle. Set sample format. Sample rate convertions
   and integer byte order swaps are left for SDL to do. */
static int
add_handle_mpg123(stream_mpg123 *s)
{
    const long *rates;
    size_t i;
    int err;
   
    s->decoder = mpg123_new(NULL, &err);
    if (err != MPG123_OK) {
        Mix_SetError(mpg123_plain_strerror(err));
        s->playback = STATE_ERROR;
        return -1;
    }

    if (mpg123_format_none(s->decoder) != MPG123_OK) {
        set_mpg123_error(s, "");
        mpg123_delete(s->decoder);
        s->decoder = NULL;
        return -1;
    }

    /* Allow all sample rates. Let SDL decide what to accept or reject. */
    for (mpg123_rates(&rates, &i); i != 0; --i) {
        if (mpg123_format(s->decoder,
                          rates[i - 1],
                          target.mpg123_channels,
                          target.mpg123_encoding) != MPG123_OK) {
            set_mpg123_error(s, "");
            mpg123_delete(s->decoder);
            s->decoder = NULL;
            return -1;
        }
    }

    /* Set up to read a RWops stream. Needs mpg123_open_handle() to open. */
    if (mpg123_replace_reader_handle(s->decoder,
                                     read_callback,
                                     seek_callback,
                                     NULL) != MPG123_OK) {
        set_mpg123_error(s, "");
        mpg123_delete(s->decoder);
        s->decoder = NULL;
        return -1;
    }

    /* No output to stderr. */
    if (mpg123_param(s->decoder,
                     MPG123_ADD_FLAGS,
                     MPG123_QUIET, 0.0) != MPG123_OK) {
        set_mpg123_error(s, "");
        mpg123_delete(s->decoder);
        s->decoder = NULL;
        return -1;
    }

    return 0;
}

/* Translate a SDL format code into a libmpg123 encoding code. */
static int
SDL_encoding_as_mpg123(int format)
{
    switch (format) {

    case AUDIO_U8:
        return MPG123_ENC_UNSIGNED_8;
    case AUDIO_S8:
        return MPG123_ENC_SIGNED_8;
    case AUDIO_U16LSB:
    case AUDIO_U16MSB:
        return MPG123_ENC_UNSIGNED_16;
    case AUDIO_S16LSB:
    case AUDIO_S16MSB:
        return MPG123_ENC_SIGNED_16;
    }

    SDL_SetError("Unsupported MP3 sample format");
    return 0;
}

/* Fill the sample buffer with properly formatted samples. If
   end-of-stream is reached, change the stream state to STATE_DONE.
   Return the number of bytes in the buffer or -1 on an error. */
static int
add_samples(stream_mpg123 *s)
{
    size_t len_read = 0;
    int code;
   
    /* Assumptions: stream s is open. If active it is also playing. */

    if (s->playback != STATE_ACTIVE) {
        /* Do not come back unless the stream is active. */
        stop_mpg123(s);
        return 0;
    }

    s->buffer_start = s->cvt.buf;
    code = mpg123_read(s->decoder,
                       s->cvt.buf,
                       s->read_block_size,
                       &len_read);

    if (code == MPG123_NEW_FORMAT) {
        code = set_convertion(s) ? MPG123_ERR : MPG123_OK;
    }
    else if (code == MPG123_OK || code == MPG123_DONE) {
        s->cvt.len = (int)len_read;
        if (SDL_ConvertAudio(&s->cvt)) {
            clear_buffer(s);
            s->playback = STATE_ERROR;
            code = MPG123_ERR;
        }
        else {
            s->buffer_len = s->cvt.len_cvt;
        }
        if (code == MPG123_DONE) {
            s->playback = STATE_DONE;
            code = MPG123_OK;
            if (s->buffer_len == 0) {
                /* The MP3 stream has finished and the samples buffer is empty.
                   Why continue? */
                stop_mpg123(s);
            }
        }
    }
    else  {
        /* mpg123_read() raised an error. */
        set_mpg123_error(s, "");
    }

    return (code == MPG123_OK) ? s->buffer_len : -1;
}
    
/* Copy up to len bytes of samples from the sample buffer of s to buffer.
   Return the number of bytes copied. Adjust for the volume is necessary.
   Uncopied samples will be copied is subsequent calls to move_samples().
   This function always succeeds. */
static int
move_samples(stream_mpg123 *s, Uint8 *buffer, int len)
{
    size_t size = s->buffer_len;

    /* Assumptions: s is an open stream, len is positive,
       buffer is not NULL and at least len bytes are allocated. */

    if (size > len) {
        size = len;
    }

    if (s->volume == MIX_MAX_VOLUME) {
        memcpy(buffer, s->buffer_start, size);
    }
    else {
        SDL_MixAudio(buffer, s->buffer_start, size, s->volume);
    }

    s->buffer_start += size;
    s->buffer_len -= size;

    return size;
}

static size_t
get_buffered_len(stream_mpg123 *s)
{
    /* Assumptions: s is an open stream. */

    return s->buffer_len;
}

static void
clear_buffer(stream_mpg123 *s)
{
    s->buffer_start = s->cvt.buf;
    s->buffer_len = 0;
}

/* Fill the streams SDL_AudioCVT field. SDL is responsible for byte swapping
   of samples, frequency conversion, and adding additional channels beyond the
   first two. */
static int
set_convertion(stream_mpg123 *s)
{
    long rate;
    int channels;
    int encoding;
    Uint16 format;
    Uint8 *buf = s->cvt.buf;
    int err;

    err = mpg123_getformat(s->decoder, &rate, &channels, &encoding);
    if (err != MPG123_OK) {
        set_mpg123_error(s, "");
        return -1;
    }
    switch (encoding) {

    case MPG123_ENC_UNSIGNED_8:
        format = AUDIO_U8;
        break;
    case MPG123_ENC_SIGNED_8:
        format = AUDIO_S8;
        break;
    case MPG123_ENC_UNSIGNED_16:
        format = AUDIO_U16SYS;
        break;
    case MPG123_ENC_SIGNED_16:
        format = AUDIO_S16SYS;
        break;
    default:
        Mix_SetError("unknown MP3 sample encoding");
        return -1;
    }
    if(SDL_BuildAudioCVT(&s->cvt,
                         format,
                         (Uint8)channels,
                         (int)rate,
                         target.mix_format,
                         target.mix_channels,
                         target.mix_freq      ) < 0) {
        s->playback = STATE_ERROR;
        return -1;
    }
    /* Don't assume SDL_BuildAudioCVT leaves the following fields untouched. */
    s->cvt.buf = buf;

    return adjust_buffer(s);
}

/* Allocate the initial sample buffer, or realloc for a new
   source stream format */
static int
adjust_buffer(stream_mpg123 *s)
{
    size_t required_size = s->read_block_size * s->cvt.len_mult;
    Uint8 *buf;

    if (required_size > s->buffer_alloc_size) {
        buf = (Uint8 *)SDL_realloc(s->cvt.buf, required_size);
        if (!buf) {
            SDL_OutOfMemory();
            s->playback = STATE_ERROR;
            return -1;
        }
        s->cvt.buf = buf;
        s->buffer_alloc_size = required_size;
    }
    return 0;
}

static void
set_mpg123_error(stream_mpg123 *s, const char *prefix)
{
    const char *msg_mpg123 = mpg123_strerror(s->decoder);

    stop_mpg123(s);
    clear_buffer(s);
    Mix_SetError("%s%s", prefix, msg_mpg123);
    s->playback = STATE_ERROR;
}

#endif /* MP3_MPG123_MUSIC */
