/*
  This file is part of a patch to SDL_mixer 1.2 to support MP3 playback
  using libmpg123. It is part of the Pygame project (www.pygame.org).

  pygame - Python Game Library
  Copyright (C) 2000-2001  Pete Shinners

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Lenard Lindstrom
  len-l@telus.net


  SDL_mixer:  An audio mixer library based on the SDL library
  Copyright (C) 1997-2012 Sam Lantinga <slouken@libsdl.org>
*/

/*
  MP3 (MPEG-1 or MPEG-2 layer III) playback using libmpg123 (http://mpg123.org)
  for decoding. The music_mp3_mpg123 C module is an addition to the SDL-mixer
  music subsystem. These functions are accessed in music.c.
*/

/* $Id$ */

#if defined(MP3_MPG123_MUSIC) && !defined(MUSIC_MP3_MPG123_H)

#define MUSIC_MP3_MPG123_H

/* A single MP3 stream. This is an opaque structure. All access is through
   module function calls. */
typedef struct stream_mpg123_s stream_mpg123;

/* Register the audio format passed to Mix_OpenAudio(). Return 0 on success,
   -1 on an error. A mixer error message is also set. The only cause for
   an error is if the requested sample format is unsupported by libmpg123.
  
   Call this function first.
 */
extern int init_mpg123(SDL_AudioSpec *output);

/* Exit the underlying mpg123 library.
 
   Call at program completion.
 */
extern void quit_mpg123(void);

/* Return a new stream that reads MP3 encoded audio and returns it
   decoded to the raw format set by the init_mpg123() call. Set freerw
   to 1 if close_mpg123() is to close the rw stream as well, otherwise
   set to 0. Argument rw must be an open RWops stream with
   read and seek support.
   
   On failure a mixer error message is set and NULL is returned.
  
   The new stream is stopped: playing_mpg123() returns 0 and read_mpg123()
   returns no samples. The initial volume is MIX_MAX_VOLUME. */
extern stream_mpg123 *open_RW_mpg123(SDL_RWops *rw, int freerw);

/* Release allocated memory and close the associated RWops stream if required.
   The pointer s becomes invalid. Argument s must be an open stream returned. */
extern void close_mpg123(stream_mpg123 *s);

/* Mark stream s as playing if it is not completed or stopped by an error.
   To start a completed stream reposition first. A stream interrupted by an
   error cannot be restarted; close it. Argument s must be an open stream. */
extern void play_mpg123(stream_mpg123 *s);

/* Return 1 if stream s is playing, 0 if stopped */
extern int playing_mpg123(stream_mpg123 *s);

/* Mark stream s as stopeed. Reading from the stream will return no samples.
   The stream position remains unchanged. Argument s must be an open stream. */
extern void stop_mpg123(stream_mpg123 *s);

/* Set the stream volume. This takes effect immediately. The next read will
   return samples at the new volume. The argument volume must be between
   0 and MIX_MAX_VOLUME inclusive. The argument s must be an open stream. */
extern void setvolume_mpg123(stream_mpg123 *s, int volume);

/* If time is positive, jump forward in the stream by that many seconds. If
   time is 0.0, rewind to the beginning. A negative time is ignored. Moving
   past end-of-stream stops playback at the next read. */
extern void position_mpg123(stream_mpg123 *s, double time);

/* Read up to len bytes of samples into buffer. Return len less the number
   of bytes actually read. If already at the end-of-stream then playback
   is stopped. Argument s must be an open stream. Argument len must be
   non-negative.

   The samples are formatted according to the audio specification passed
   to init_mpg123(). Sample values are scaled to the current volume set
   for the stream.

   If an error occurs, the stream is stopped, no samples are returned, and
   a mixer error message is set. */
extern int read_mpg123(stream_mpg123 *s, Uint8 *buffer, int len);

#endif /* defined(MP3_MPG123_MUSIC) && !defined(MUSIC_MP3_MPG123_H) */
