#!/usr/bin/env python
# -*- coding: ascii -*-
# Program build_on_msys.py
# Requires Python 2.7 or later and win32api.

"""Build Pygame dependencies using MinGW-W64 and MSYS2

Configured for Pygame 1.9.2 and Python 2.7 or higher.

By default the libraries are installed in ..\build-x64 (or ..\build-x86)
unless a different directory is specified by the --prefix command line
argument.

This program can be run from a Windows cmd.exe or a MSYS2 terminal. The current
directory and its outer directory are searched for the library source
directories. Run the program from the pygame trunk directory. The Windows
file path cannot have spaces in it.

The recognized, and optional, environment variables are:
  PREFIX - Destination directory
  MSYS_ROOT_DIRECTORY - MSYS home directory
  CPPFLAGS - preprocessor options, appended to options set by the program
  CFLAGS - compiler flags, appended to options set by the program
  LDFLAGS - linker options, prepended to flags set by the program

To get a list of command line options run

python build_deps.py --help

This program has been tested against the following libraries:

SDL 1.2.15+ (rev 9f8a2b28f61d)
SDL_image 1.2.13 (rev 0c88c81b59c9)
SDL_mixer 1.2.13 (rev a4e9c53d9c30) patched for Pygame
SDL_ttf 2.0.12 (rev 1ae8bba8f1d5)
freetype 2.4.12
libogg 1.3.2
libvorbis 1.3.5
FLAC 1.3.1
mikmod 3.1.12 patched (included with SDL_mixer 1.2.13)
libmpg123 1.22.4
tiff 4.0.5
libpng 1.6.18
jpeg 8d
zlib 1.2.8
PortMidi revision 217

The build environment used: 

MinGW-W64 gcc 4.9.3 installed as C:\mingw-w64
MSYS2 2.2.1 installed in C:\msys32 with additional packages:
  autoconf 2.69-3
  automake-wrapper 10-1
  bash-completion 2.1-5
  bsdcpio 3.1.2-5
  bsdtar 3.1.2-5
  catgets 1.1-2
  crypt 1.1-6
  file 5.24-1
  flex 2.5.39-4
  inetutils 1.9.2-1
  libtool 2.4.6-1
  lndir 1.0.3-1
  make 4.1-4
  mintty 1~2.0.3-1
  patch 2.7.5-1
  pkgfile 15-1
  rebase 4.4.1-6
  repman-git r23.87bf865-1
  time 1.7-1
  tzcode 2015.e-1
  unzip 6.0-2
  util-linux 2.26.2-1
nasm 2.11.08

The libraries are linked to msvcrt.dll, the VC 6.0 runtime library.

The build has been performed on Windows XP, SP3.

Build issues:
  An intermittent problem was noted with SDL's configure involving locking of
  conftest.exe resulting in various C library functions being reported unavailable
  when in fact they are present. This does not appear to be a problem with the
  configure script itself but rather Msys. If it happens then just rerun
  build_on_msys.py.
"""

import msys

from optparse import OptionParser
import os
import sys
from glob import glob
import time

# For Python 2.x/3.x compatibility
def geterror():
    return sys.exc_info()[1]

#
#   Generic declarations
#
hunt_paths = ['.', '..']

parent_dir_wp = os.path.dirname(os.getcwd())

def prompt(p=None):
    """MSYS friendly raw_input

    This provides a hook that can be replaced for testing.
    """

    msys.msys_raw_input(p)

def print_(*args, **kwds):
    msys.msys_print(*args, **kwds)

def confirm(message):
    """Ask a yes/no question, return result"""

    reply = prompt("\n%s [Y/n]:" % message)
    if reply and reply[0].lower() == 'n':
        return 0
    return 1

def as_flag(b):
    """Return bool b as a shell script flag '1' or '0'"""

    if b:
        return '1'
    return '0'

def as_linker_lib_path(p):
    """Return as an ld library path argument"""

    if p:
        return '-L' + p
    return ''

def as_linker_option(p):
    """Return as an ld library path argument"""

    if p:
        return '-Wl,' + p
    return ''

def as_preprocessor_header_path(p):
    """Return as a C preprocessor header include path argument"""

    if p:
        return '-I' + p
    return ''

def as_macro_define(m, v):
    """Return as a C preprocessor command line macro definition"""

    if v:
        return '-D%s=%s' % (m, v)
    return '-D%s' % (m,)

def merge_strings(*args, **kwds):
    """Return non empty strings joined by sep

    The default separator is an empty string.
    """

    sep = kwds.get('sep', '')
    return sep.join([s for s in args if s])

class BuildError(Exception):
    """Raised for missing source paths and failed script runs"""
    pass

def get_ptr_size():
    return 64 if sys.maxsize > 2**32 else 32

def as_target_triplet(size):
    """Return pointer bit size (32 or 64) as a GNU target triplet"""
    if size == 32:
        return 'i686-w64-mingw32'
    if size == 64:
        return 'x86_64-w64-mingw32'
    raise BuildError("Unknown pointer size {}".format(size))

def as_target_type(size):
    """Return pointer bit size as a MinGW-W64 gcc build option"""
    return "-m{:d}".format(size)

def as_machine_type(size):
    """Return pointer bit size as a Windows machine type"""
    if size == 32:
        return "x86"
    if size == 64:
        return "x64"
    raise BuildError("Unknown pointer size {}".format(size))

default_machine_type = as_machine_type(get_ptr_size())

class Dependency(object):
    """Builds a library"""

    def __init__(self, name, wildcards, libs, shell_script):
        self.name = name
        self.wildcards = wildcards
        self.shell_script = shell_script
        self.libs = libs

    def configure(self, hunt_paths):
        self.path = None
        self.paths = []
        self.hunt(hunt_paths)
        self.choosepath()

    def hunt(self, hunt_paths):
        parent = os.path.abspath('..')
        for p in hunt_paths:
            for w in self.wildcards:
                found = glob(os.path.join(p, w))
                found.sort() or found.reverse()  #reverse sort
                for f in found:
                    if f[:5] == '..'+os.sep+'..' and \
                        os.path.abspath(f)[:len(parent)] == parent:
                        continue
                    if os.path.isdir(f):
                        self.paths.append(f)

    def choosepath(self):
        path = None
        if not self.paths:
            raise BuildError("Path for %s: not found" % self.name)
        if len(self.paths) == 1:
            path = self.paths[0]
        else:
            print_("Select path for %s:" % self.name)
            for i in range(len(self.paths)):
                print_("  %d = %s" % (i+1, self.paths[i]))
            print_("  0 = <Nothing>")
            choice = prompt("Select 0-%d (1=default):" % len(self.paths))
            if not choice:
                choice = 1
            else:
                choice = int(choice)
            if choice > 0:
                path = self.paths[choice-1]
        if path is not None:
            self.path = os.path.abspath(path)

    def build(self, msys):
        if self.path is not None:
            msys.environ['BDDLLS'] = merge_strings(*self.libs, sep=' ')
            return_code = msys.run_shell_script(self.shell_script, self.path)
            if return_code != 0:
                raise BuildError("The build for %s failed with code %d" %
                                 (self.name, return_code))
        else:
            raise BuildError("No source directory for %s" % self.name)

class Preparation(object):
    """Perform necessary build environment preparations"""

    def __init__(self, name, shell_script):
        self.name = name
        self.path = '.'
        self.paths = []
        self.libs = []
        self.shell_script = shell_script

    def configure(self, hunt_paths):
        pass

    def build(self, msys):
        return_code = msys.run_shell_script(self.shell_script, self.path)
        if return_code != 0:
            raise BuildError("Preparation '%s' failed with code %d" %
                             (self.name, return_code))

def configure(dependencies, hunt_paths):
    """Find source directories of all dependencies"""

    success = True
    print_("Hunting for source directories...")
    for dep in dependencies:
        try:
            dep.configure(hunt_paths)
        except BuildError:
            print_(geterror())
            success = False
        else:
            if dep.path:
                print_("Source directory for", dep.name, ":", dep.path)
    if not success:
        raise BuildError("Not all source directories were found")

def build(dependencies, msys):
    """Execute the shell scripts for all dependencies"""

    for dep in dependencies:
        print_("\n\n----", dep.name, "----")
        dep.build(msys)

def command_line():
    """Process the command line and return the options"""

    usage = ("usage: %prog [options] --all\n"
             "       %prog [options] [args]\n"
             "\n"
             "Build the Pygame dependencies. The args, if given, are\n"
             "libraries to include or exclude.\n"
             "\n"
             "At startup this program may prompt for missing information.\n"
             "Be aware of this before redirecting output or leaving the\n"
             "program unattended. Once the 'Starting build' message appears\n"
             "no more user input is required. The build process will\n"
             "abort on the first error, as library build order is important.\n"
             "\n"
             "See --help-args.\n"
             "\n"
             "For more details see the program's document string\n")

    parser = OptionParser(usage)
    parser.add_option('-a', '--all', action='store_true', dest='build_all',
                      help="Include all libraries in the build")
    parser.set_defaults(build_all=False)
    parser.add_option('--no-configure', action='store_false', dest='configure',
                      help="Do not prepare the makefiles")
    parser.set_defaults(configure=True)
    parser.add_option('--no-compile', action='store_false', dest='compile',
                      help="Do not compile or install the libraries")
    parser.set_defaults(compile=True)
    parser.add_option('--no-install', action='store_false', dest='install',
                      help="Do not install the libraries")
    parser.add_option('--no-strip', action='store_false', dest='strip',
                      help="Do not strip the library")
    parser.set_defaults(strip=True)
    parser.set_defaults(install=True)
    parser.add_option('--clean', action='store_true', dest='clean',
                      help="Remove generated files (make clean)"
                           " as a last step")
    parser.set_defaults(clean=False)
    parser.add_option('--clean-only', action='store_true', dest='clean_only',
                      help="Perform only a clean")
    parser.set_defaults(clean_only=False)
    parser.add_option('-e', '--exclude', action='store_true', dest='exclude',
                      help="Exclude the specified libraries")
    parser.set_defaults(exclude=False)
    parser.add_option('-m', '--msys-root', action='store',
                      dest='msys_directory',
                      help="MSYS directory path, which may include"
                           " the 1.x subdirectory")
    parser.set_defaults(msys_directory='')
    parser.add_option('-s', '--sources', action='store',
                      dest='sources',
                      help="Paths to search for library source directories"
                           " as a semicolon ';' separated list: defaults to %s"
                           % (';'.join(hunt_paths),))
    parser.add_option('-p', '--prefix', action='store',
                      dest='prefix_wp',
                      help="Destination directory of the build")
    parser.set_defaults(prefix_wp='')
    parser.add_option('--help-args', action='store_true', dest='arg_help',
                      help="Show a list of recognised libraries,"
                           " in build order, and exit")
    parser.set_defaults(arg_help=False)
    parser.add_option('--subsystem-noforce', action='store_true', dest='subsystem_noforce',
                      help="Do not force the dlls to build with the GUI subsystem type")
    parser.set_defaults(subsystem_noforce=False)
    parser.add_option('-b', '--beep', action='store_true', dest='finish_alert',
                      help="Beep the computer speaker when finished.")
    parser.set_defaults(finish_alert=False)
    parser.add_option('-n', '--beep-ntimes', type='int', action='store', dest='finish_alert_ntimes',
                      help="Beep the computer speaker n times when finished")
    parser.set_defaults(finish_alert_ntimes=0)
    parser.add_option('-t', '--target', type='choice', choices=['x86', 'x64'],
                      dest='target_type',
                      help="Build for x86 or x64 Windows")
    parser.set_defaults(target_type=default_machine_type)
    return parser.parse_args()

def set_environment_variables(msys, options):
    """Set the environment variables used by the scripts"""

    environ = msys.environ

    msys_root_wp = msys.msys_root
    prefix_wp = options.prefix_wp
    if not prefix_wp:
        prefix_wp = environ.get('PREFIX', '')
    if prefix_wp:
        prefix_mp = msys.windows_to_msys(prefix_wp)
    else:
        prefix_wp = os.path.join(parent_dir_wp,
                                 "build-" + options.target_type)
    prefix_mp = msys.windows_to_msys(prefix_wp)
    include_mp = prefix_mp + '/include'
    lib_mp = prefix_mp + '/lib'
    subsystem = ''
    if not options.subsystem_noforce:
        subsystem = '-mwindows'

    # Explicitly spell out host and build systems for configure
    # scripts. The SDL-1.2.15 and jpeg-8d scripts do not recognize
    # 'msys' as an operating system, for instance. Also, only
    # treat the building of 64-bit binaries on a 32-bit Windows
    # as cross-compiling (host differs from build), as 32-bit
    # binaries run under 64-bit Windows.
    host_machine = options.target_type
    host_ptr_size = 64 if host_machine == 'x64' else 32
    host_system = as_target_triplet(host_ptr_size)
    build_system = as_target_triplet(get_ptr_size())

    environ['BDPROG'] = sys.argv[0]
    environ['PREFIX'] = prefix_mp
    environ['BDBUILD'] = build_system
    environ['BDHOST'] = host_system
    environ.pop('INCLUDE', None)  # INCLUDE causes problems with MIXER.
    environ['CPPFLAGS'] = merge_strings(as_preprocessor_header_path(include_mp),
                                        environ.get('CPPFLAGS', ''),
                                        sep=' ')
    environ['CFLAGS'] = merge_strings(as_target_type(host_ptr_size),
                                      environ.get('CFLAGS', ''),
                                      sep=' ')
    environ['LDFLAGS'] = merge_strings(environ.get('LDFLAGS', ''),
                                       as_target_type(host_ptr_size),
                                       as_linker_lib_path(lib_mp),
                                       subsystem,
                                       sep=' ')

    environ['BDTARGET'] = options.target_type
    environ['BDSUBDIR'] = merge_strings('pgbuild', options.target_type, sep='-')
    environ['BDCONF'] = as_flag(options.configure and
                                not options.clean_only)
    environ['BDCOMP'] = as_flag(options.compile and
                                not options.clean_only)
    environ['BDINST'] = as_flag(options.install and
                                options.compile and
                                not options.clean_only)
    environ['BDSTRIP'] = as_flag(options.compile and
                                 options.install and
                                 options.strip and
                                 not options.clean_only)
    environ['BDCLEAN'] = as_flag(options.clean or options.clean_only)
    environ['BDMSVCRT_VERSION'] = '%i' % (options.msvcrt_version,)

    # Number of alert beeps
    nbeeps = options.finish_alert_ntimes
    if options.finish_alert and nbeeps <= 0:
        nbeeps = 1
    environ['BDNBEEPS'] = str(nbeeps)

    mingw_bin_wp = os.path.join(msys.mingw_root, 'bin')
    environ['PATH'] = merge_strings(mingw_bin_wp, environ['PATH'], sep=';')

class ChooseError(Exception):
    """Failure to select dependencies"""
    pass

def choose_dependencies(dependencies, options, args):
    """Return the dependencies to actually build"""

    if options.build_all:
        if args:
            raise ChooseError("No library names are accepted"
                              " for the --all option.")
        if options.exclude:
            return []
        else:
            return dependencies

    if args:
        names = [d.name for d in dependencies]
        args = [a.upper() for a in args]
        for a in args:
            if a not in names:
                msg = ["%s is an unknown library; valid choices are:" % a]
                msg.extend(names)
                raise ChooseError('\n'.join(msg))
        if options.exclude:
            return [d for d in dependencies if d.name not in args]
        return [d for d in dependencies if d.name in args]

    return []

def summary(dependencies, msys, start_time, chosen_deps):
    """Display a summary report of new, existing and missing libraries"""

    import datetime

    print_("\n\n=== Summary ===")
    if start_time is not None:
        print_("  Elapse time:",
               datetime.timedelta(seconds=time.time()-start_time))
    print_()
    for dep in chosen_deps:
        if dep.path is None:
            print_("  ** No source directory found for", dep.name)
        elif dep.path:
            print_("  Source directory for", dep.name, ":", dep.path)
    print_()
    prefix = msys.msys_to_windows(msys.environ['PREFIX']).replace('/', os.sep)
    bin_dir = os.path.join(prefix, 'bin')
    lib_dir = os.path.join(prefix, 'lib')
    for d in dependencies:
        for lib in d.libs:
            if lib.endswith('.dll'):
                lib_path = os.path.join(bin_dir, lib)
                try:
                    mod_time = os.path.getmtime(lib_path)
                except:
                    msg = "No DLL"
                else:
                    if mod_time >= start_time:
                        msg = "Installed new DLL %s" % (lib_path,)
                    else:
                        msg = "-- (old DLL %s)" % (lib_path,)
            elif lib.endswith('.a'):
                lib_path = os.path.join(lib_dir, lib)
                try:
                    mod_time = os.path.getmtime(lib_path)
                except:
                    msg = "No static library"
                else:
                    if mod_time >= start_time:
                        msg = "Installed new static library %s" % (lib_path,)
                    else:
                        msg = "-- (old static library %s)" % (lib_path,)
            else:
                msg = "Internal error: unknown library type %s" % (lib,)
            print_("  %-10s: %s" % (d.name, msg))

def main(dependencies, msys_preparation):
    """Build the dependencies according to the command line options."""

    options, args = command_line()
    if options.arg_help:
        print_("These are the Pygame library dependencies:")
        for dep in dependencies:
            print_(" ", dep.name)
        return 0

    # Only building against msvcrt.dll for now.
    options.msvcrt_version = 60

    try:
        chosen_deps = choose_dependencies(dependencies, options, args)
    except ChooseError:
        print_(geterror())
        return 1
    if not chosen_deps:
        if not args:
            print_("No libraries specified.")
        elif options.build_all:
            print_("All libraries excluded")
    print_("Linking to C runtime library msvcrt.dll.")
    if chosen_deps and not options.clean_only:
        chosen_deps.insert(0, msys_preparation)
    try:
        m = msys.Msys(options.msys_directory)
    except msys.MsysException:
        print_(geterror())
        return 1
    print_("Using MSYS in directory:", m.msys_root)
    print_("MinGW directory:", m.mingw_root)
    start_time = None
    return_code = 1
    set_environment_variables(m, options)
    report_environment.build(m)
    sources = hunt_paths
    if options.sources:
        sources = options.sources.split(';')
    print_("library source directories search paths: %s" % (';'.join(sources),))
    try:
        configure(chosen_deps, sources)
    except BuildError:
        print_("Build aborted:", geterror())
    else:
        if options.clean_only:
            print_("\n=== Performing clean ===")
        else:
            print_("\n=== Starting build ===")
        start_time = time.time()  # For file timestamp checks.
        try:
            build(chosen_deps, m)
        except BuildError:
            print_("Build aborted:", geterror())
        else:
            # A successful build!
            return_code = 0
    if not options.clean_only:
        summary(dependencies, m, start_time, chosen_deps)

    if int(m.environ['BDNBEEPS']) > 0:
        alert.build(m)
    return return_code

#
#   Build specific code
#

# This list includes the MSYS shell scripts to build each library. Each script
# runs in an environment where MINGW_ROOT_DIRECTORY is defined and the MinGW
# bin directory is in PATH. The GNU configure script build and host
# options are available as BDBUILD and BDBUILD. BDTARGET is "x86" for a
# 32-bit build, "x64" for a 64-bit build. Four build control environment variables
# are defined: BDCONF, BDCOMP, BDINST and BDCLEAN. They are either '0' or '1'.
# They represent configure, compile, install and clean respectively.
# When '1' the corresponding action is performed. When '0' it is skipped.
# The installation directory is given by PREFIX. The script needs to prepend
# it to PATH. The script's HOME directory is the source code root directory.
# Various gcc flags are in CPPFLAGS, CFLAGS, and LDFLAGS. INCLUDE is undefined.
#
# None of these scripts end with an "exit". Exit, possibly, leads to Msys
# freezing on some versions of Windows (98).
# 
# The list order corresponds to build order. It is critical.
dependencies = [
    Dependency('SDL', ['SDL-9f8a2b28f61d'], ['SDL.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  # If this comes from the repository it has no configure script
  if [ ! -f "./configure" ]; then
    ./autogen.sh
  fi

  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR

  cd $BDSUBDIR
  ../configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
               --disable-static --disable-stdio-redirect \
               CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS" \

  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi
  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  case $BDTARGET in
    x86) windres="windres --target=pe-i386";;
    x64) windres="windres --target=pe-x86-64";;
    *) echo Unknown target format $BDTARGET; exit 1;;
  esac
  make WINDRES="$windres"
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install-bin install-hdrs install-lib
  # Need SDL_config.h.default and SDL_config_win32.h for Windows.
  cp -f "$HOME/include/SDL_config_win32.h" "$PREFIX/include/SDL"
  cp -f "$HOME/include/SDL_config.h.default" "$PREFIX/include/SDL/SDL_config.h"
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [ x$BDCLEAN == x1 ]; then
  cd $BDSUBDIR
  set +e
  make clean
  cd ..
fi
"""),
    Dependency('Z', ['zlib-[1-9].*'], ['zlib1.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR
  cp *.c $BDSUBDIR
  cp *.h $BDSUBDIR
  cp zlib.pc.in $BDSUBDIR
  cp -a "$PWD/win32" $BDSUBDIR/win32
  cp -a "$PWD/test" $BDSUBDIR/test
  cp -fp win32/Makefile.gcc $BDSUBDIR
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  case $BDTARGET in
    x86) windres="windres --target=pe-i386";;
    x64) windres="windres --target=pe-x86-64";;
    *) echo Unknown target format $BDTARGET; exit 1;;
  esac
  # Build with the import library renamed, using asm code, our CPPFLAGS,
  # CFLAGS, and LDFLAGS (passed in as LOC).
  make IMPLIB=libz.dll.a -fMakefile.gcc \
       SHARED_MODE=1  RC="$windres"\
       CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LOC="$LDFLAGS"
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  # Make sure everything is installed in the correct places
  make install LIBRARY_PATH="$PREFIX/lib" INCLUDE_PATH="$PREFIX/include" \
    BINARY_PATH="$PREFIX/bin" SHARED_MODE=1 IMPLIB=libz.dll.a -fMakefile.gcc
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [ x$BDCLEAN == x1 ]; then
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('FREETYPE', ['freetype-[2-9].*'], ['libfreetype-6.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR

  cd $BDSUBDIR
  ../configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
               --disable-static \
               CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS"

  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' builds/unix/config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi
  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  # Compile the apinames.c build tool as a 32-bit program.
  make CCexe="gcc -m32"
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [ x$BDCLEAN == x1 ]; then
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('FONT', ['SDL_ttf-1ae8bba8f1d5'], ['SDL_ttf.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  # If this comes from the repository it has no configure script
  if [ ! -f "./configure" ]; then
    ./autogen.sh
  fi

  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR

  cd $BDSUBDIR
  ../configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
               --disable-static \
               --with-freetype-prefix="$PREFIX" --with-sdl-prefix="$PREFIX" \
               CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS"

  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi
  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  case $BDTARGET in
    x86) windres="windres --target=pe-i386";;
    x64) windres="windres --target=pe-x86-64";;
    *) echo Unknown target format $BDTARGET; exit 1;;
  esac
  make WINDRES="$windres"
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [ x$BDCLEAN == x1 ]; then
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('PNG', ['l*png*[1-9][1-9.]*'], ['libpng16-16.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR

  cd $BDSUBDIR
  ../configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
               --disable-static \
               CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS"

  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi
  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  make
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [ x$BDCLEAN == x1 ]; then
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('JPEG', ['jpeg-8d'], ['libjpeg-8.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR

  cd $BDSUBDIR

  ../configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
               --disable-static \
               CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS"

  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi

  cp ../jconfig.vc jconfig.h
  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  make
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [ x$BDCLEAN == x1 ]; then
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('TIFF', ['tiff-[3-9].*'], ['libtiff-5.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

# Only building the library.
bd_subdirs="port libtiff"


if [ x$BDCONF == x1 ]; then
  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR

  cd $BDSUBDIR
  ../configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
               --disable-static --disable-cxx \
               CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS"

  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi
  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  make SUBDIRS="$bd_subdirs"
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install SUBDIRS="$bd_subdirs"
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [ x$BDCLEAN == x1 ]; then
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('WEBP', ['libwebp-[0].*'], ['libwebp-5.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  # If this comes from the repository it has no configure script
  if [ ! -f "./configure" ]; then
    ./autogen.sh
  fi

  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR

  cd $BDSUBDIR
  ../configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
               --disable-static \
               CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS"

  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi
  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  make
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [ x$BDCLEAN == x1 ]; then
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('IMAGE', ['SDL_image-0c88c81b59c9'], ['SDL_image.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  # If this comes from the repository it has no configure script
  if [ ! -f configure ]; then
    ./autogen.sh
  fi

  # configure searches for the JPEG dll. Unfortunately it uses the wrong file
  # name. Correct this.
  if [ ! -f configure.orig ]; then
    cp configure configure.orig
    sed -e 's|jpeg\.dll|libjpeg-*.dll|' configure.orig >configure
    chmod u+x configure
  fi

  # Add the destination bin directory to the library search path so
  # configure can find its precious DLL files.
  export LDFLAGS="$LDFLAGS -L$PREFIX/bin"

  # Add path to PNG headers
  CPPFLAGS="$CPPFLAGS `$PREFIX/bin/libpng-config --I_opts`"

  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR

  cd $BDSUBDIR
  # Disable dynamic loading of image libraries as it uses the wrong DLL
  # search path: does not check in the same directory.
  #  --disable-libtool-lock: Prevent libtool deadlocks (maybe).
  ../configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
               --disable-static --disable-libtool-lock \
               --disable-jpg-shared --disable-png-shared \
               --disable-tif-shared --disable-webp-shared \
               CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS"

  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi
  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  case $BDTARGET in
    x86) windres="windres --target=pe-i386";;
    x64) windres="windres --target=pe-x86-64";;
    *) echo Unknown target format $BDTARGET; exit 1;;
  esac
  make WINDRES="$windres"
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [[ x$BDCLEAN == x1 && -f Makefile ]]; then
  if [ -f configure.orig ]; then
    rm -f configure
    mv configure.orig configure
  fi
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('OGG', ['libogg-[1-9].*'], ['libogg-0.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR

  cd $BDSUBDIR
  ../configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
               --disable-static \
               CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS"

  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi
  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  make
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [[ x$BDCLEAN == x1 && -f Makefile ]]; then
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('VORBIS',
               ['libvorbis-[1-9].*'],
               ['libvorbis-0.dll', 'libvorbisfile-3.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR

  cd $BDSUBDIR
  # configure looks for a previous build. Remove if found.
  if [ -f "Makefile" ]; then
    make distclean
  fi
  ../configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
               --disable-static \
               CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS" LIBS='-logg'
  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi
  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  make
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [[ x$BDCLEAN == x1 && -f Makefile ]]; then
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('MPG123', ['mpg123-[1-9].*'], ['libmpg123-0.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR

  cd $BDSUBDIR
  ../configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
               CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS"

  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi
  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  make
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [[ x$BDCLEAN == x1 && -f Makefile ]]; then
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('FLAC', ['flac-[1-9].*'], ['libFLAC.a'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  rm -rf $BDSUBDIR
  workdir=$(mktemp -d --tmpdir pygame.XXXXXXXX)
  cp -a . $workdir
  mv $workdir $BDSUBDIR

  cd $BDSUBDIR

  # Add __MINGW32__ to SIZE_T_MAX declaration test in alloc.h header.
  BDHDR='include/share/alloc.h'
  BDTMP='alloc.h_'
  cp -f "$BDHDR" "$BDTMP"
  sed 's/^#  ifdef _MSC_VER$/#  if defined _MSC_VER || defined __MINGW32__/' \
    "$BDTMP" >"$BDHDR"
  rm "$BDTMP"

  # Will only install a static library, but that is all that is needed.
  ./configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
              --disable-shared --disable-ogg --disable-cpplibs \
              --disable-doxygen-docs \
              CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS"

  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi

  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  make
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cp $BDSUBDIR/src/libFLAC/.libs/libFLAC.a "$PREFIX/lib"
  mkdir -p "$PREFIX/include/FLAC"
  cp -f $BDSUBDIR/include/FLAC/*.h "$PREFIX/include/FLAC"
fi

if [[ x$BDCLEAN == x1 ]]; then
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('MIKMOD', ['libmikmod-3.*'], ['libmikmod.a'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  # As the source code will need modification, first make a working copy of
  # this directory and move to ./$BDSUBDIR
  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR
  cp -r -t $BDSUBDIR \
     include drivers loaders playercode mmio win32

  # Remove ULONG and BOOL declarations.
  #
  # These are already defined by MinGW.
  sed -i -e 96,97d -e 107,110d $BDSUBDIR/include/mikmod.h.in

  # Replace '_mm_' prefix with '_mikmod_' in source code.
  #
  # This is painful, MikMod defines _mm_malloc and _mm_free, which
  # conflict with definitions in mingw-w64 definitions in malloc.h.
  # So let us give all '_mm_' identifiers safe names in every.c and .h file.
  find $BDSUBDIR -type f -name '*.[ch]' -execdir \
    sed -i -e 's/_mm_\\([a-zA-Z_]\\+\\)/_mikmod_\\1/g' '{}' '+'

  function mikmod_getver
  {
    gawk '\
      function getnum(s)
      {
        match(s, /[0-9]+/)
        return substr(s, RSTART, RLENGTH)
      }
      /^LIBMIKMOD_MAJOR_VERSION *= *[0-9]+/ { major = getnum($0); next}
      /^LIBMIKMOD_MINOR_VERSION *= *[0-9]+/ { minor = getnum($0); next}
      /^LIBMIKMOD_MICRO_VERSION *= *[0-9]+/ { micro = getnum($0); next}
      END { printf "%s %s %s", major, minor, micro }' \
      $1
    }
  export mikmod_version=($(mikmod_getver configure.in))

  sed -e "s~@prefix@~$PREFIX~g" \
      -e "s~@exec_prefix@~$PREFIX~g" \
      -e "s~@LIBMIKMOD_VERSION@~${mikmod_version[0]}.${mikmod_version[1]}.${mikmod_version[2]}~g" \
      -e "s~@REENTRANT@~-D_REENTRANT~g" \
      -e "s~@LIB_LDADD@~~g" \
      -e "s~@LIBRARY_LIB@~-lpthread~g" \
      libmikmod-config.in >$BDSUBDIR/libmikmod-config
  chmod u+x $BDSUBDIR/libmikmod-config

  # Replace _DLL with __DLL__ in macro check.
  #
  # MinGW-w64 always defines _DLL, even for static library builds.
  sed -i -e 's/\\([^a-zA-Z0-9_]\\)_DLL\\([^a-zA-Z0-9_]\\?\\)/\\1__DLL__\\2/g' \
      $BDSUBDIR/include/mikmod.h.in

  # Do not define strcasecmp macro. MinGW-w64 already does.
  #
  sed -i -e '40s/||defined(WIN32) *$//' $BDSUBDIR/include/mikmod_internals.h

  # Enviroment variable expansion is used in writing the makefile...
  cat > $BDSUBDIR/win32/Makefile.static.mingw << THE_END
# MinGW Makefile adapted from template for use under win32
#
# Automatically generated by $BDPROG
# at $(date).
#
# libmikmod subdirectory

# Your compiler here
CC=gcc
# Compiler flags
CPPFLAGS_MIKMOD=-c -DWIN32 -DDRV_DS -DDRV_WIN -DHAVE_FCNTL_H -DHAVE_MALLOC_H -DHAVE_LIMITS_H $CPPFLAGS
COMPILE=\\$(CC) \\$(CPPFLAGS_MIKMOD) -I../include -I.. -I../win32 $CFLAGS

.SUFFIXES:
.SUFFIXES: .o .c

LIBNAME=libmikmod.a

LIBS=\\$(LIBNAME)

DRIVER_OBJ=drv_ds.o drv_win.o

OBJ=\\$(DRIVER_OBJ) \\
    drv_nos.o drv_raw.o drv_stdout.o drv_wav.o \\
    load_669.o load_amf.o load_dsm.o load_far.o load_gdm.o load_it.o  \\
    load_imf.o load_m15.o load_med.o load_mod.o load_mtm.o load_okt.o \\
    load_s3m.o load_stm.o load_stx.o load_ult.o load_uni.o load_xm.o \\
    mmalloc.o mmerror.o mmio.o \\
    mdriver.o mdreg.o mloader.o mlreg.o mlutil.o mplayer.o munitrk.o mwav.o \\
    npertab.o sloader.o virtch.o virtch2.o virtch_common.o

all:            \\$(LIBS)

clean:
\tfor f in \\$(LIBS) ; do rm -f $f; done
\trm -f *.o
\trm -f mikmod_build.h

distclean:
\trm -f ../include/mikmod.h

install:
\tcp -fp libmikmod.a "\\$(PREFIX)/lib"
\tcp -fp ../include/mikmod.h "\\$(PREFIX)/include"
\tcp -fp ../libmikmod-config "\\$(PREFIX)/bin"

\\$(LIBNAME):     \\$(OBJ)
\tar -r \\$(LIBNAME) *.o
\tranlib \\$(LIBNAME)

../include/mikmod.h ../win32/mikmod_build.h:\t../include/mikmod.h.in
\tsed -e "s~@LIBMIKMOD_MAJOR_VERSION@~${mikmod_version[0]}~" \\
\t    -e "s~@LIBMIKMOD_MINOR_VERSION@~${mikmod_version[1]}~" \\
\t    -e "s~@LIBMIKMOD_MICRO_VERSION@~${mikmod_version[2]}~" \\
\t    -e "s~@DOES_NOT_HAVE_SIGNED@~~" \\
\t    ../include/mikmod.h.in >../win32/mikmod_build.h
\tcp -f ../win32/mikmod_build.h ../include/mikmod.h

drv_ds.o:       ../drivers/drv_ds.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../drivers/drv_ds.c
drv_nos.o:      ../drivers/drv_nos.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../drivers/drv_nos.c
drv_raw.o:      ../drivers/drv_raw.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../drivers/drv_raw.c
drv_stdout.o:   ../drivers/drv_stdout.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../drivers/drv_stdout.c
drv_wav.o:      ../drivers/drv_wav.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../drivers/drv_wav.c
drv_win.o:       ../drivers/drv_win.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../drivers/drv_win.c
load_669.o:     ../loaders/load_669.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_669.c
load_amf.o:     ../loaders/load_amf.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_amf.c
load_dsm.o:     ../loaders/load_dsm.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_dsm.c
load_far.o:     ../loaders/load_far.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_far.c
load_gdm.o:     ../loaders/load_gdm.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_gdm.c
load_it.o:      ../loaders/load_it.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_it.c
load_imf.o:     ../loaders/load_imf.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_imf.c
load_m15.o:     ../loaders/load_m15.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_m15.c
load_med.o:     ../loaders/load_med.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_med.c
load_mod.o:     ../loaders/load_mod.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_mod.c
load_mtm.o:     ../loaders/load_mtm.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_mtm.c
load_okt.o:     ../loaders/load_okt.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_okt.c
load_s3m.o:     ../loaders/load_s3m.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_s3m.c
load_stm.o:     ../loaders/load_stm.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_stm.c
load_stx.o:     ../loaders/load_stx.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_stx.c
load_ult.o:     ../loaders/load_ult.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_ult.c
load_uni.o:     ../loaders/load_uni.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_uni.c
load_xm.o:      ../loaders/load_xm.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../loaders/load_xm.c
mmalloc.o:      ../mmio/mmalloc.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../mmio/mmalloc.c
mmerror.o:      ../mmio/mmerror.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../mmio/mmerror.c
mmio.o:         ../mmio/mmio.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../mmio/mmio.c
mdriver.o:      ../playercode/mdriver.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/mdriver.c
mdreg.o:        ../playercode/mdreg.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/mdreg.c
mloader.o:      ../playercode/mloader.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/mloader.c
mlreg.o:        ../playercode/mlreg.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/mlreg.c
mlutil.o:       ../playercode/mlutil.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/mlutil.c
mplayer.o:      ../playercode/mplayer.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/mplayer.c
munitrk.o:      ../playercode/munitrk.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/munitrk.c
mwav.o:         ../playercode/mwav.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/mwav.c
npertab.o:      ../playercode/npertab.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/npertab.c
sloader.o:      ../playercode/sloader.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/sloader.c
virtch.o:       ../playercode/virtch.c ../playercode/virtch_common.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/virtch.c
virtch2.o:      ../playercode/virtch2.c ../playercode/virtch_common.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/virtch2.c
virtch_common.o:        ../playercode/virtch_common.c \\
\t                ../win32/mikmod_build.h ../include/mikmod_internals.h
\t\\$(COMPILE) -o \\$@ ../playercode/virtch_common.c
THE_END

    unset -v mikmod_version
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR/win32
  make -fMakefile.static.mingw
  cd ../..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR/win32
  make install -fMakefile.static.mingw
  cd ../..
fi

if [[ x$BDCLEAN == x1 && -f Makefile ]]; then
  rm -rf $BDSUBDIR
fi

"""),
    Dependency('MIXER', ['SDL_mixer-[0-9a-f]*'], ['SDL_mixer.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

mikmod_dependencies='-ldsound'
flac_dependencies='-lWs2_32'

if [ x$BDCONF == x1 ]; then
  # If this came from SVN then need a configure script.
  if [ ! -f "./configure" ]; then
    ./autogen.sh
  fi

  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR

  cd $BDSUBDIR

  # gcc supports some functions as builtins which are normally not available
  # on Windows. Use these builtins.
  export CPPFLAGS="-DHAVE_STRDUP -DHAVE_VSNPRINTF $CPPFLAGS"

  # No dynamic loading of dependent libraries. Use LIBS so FLAC test
  # builds (unfortunately LIBS is not passed on to Makefile).
  ../configure --prefix="$PREFIX" --build=$BDBUILD --host=$BDHOST \
               --disable-static --disable-libtool-lock \
               --disable-music-ogg-shared --disable-music-mod-shared \
               --disable-music-flac-shared --disable-music-fluidsynth-midi \
               --disable-static --with-sdl-prefix="$PREFIX" \
               CPPFLAGS="$CPPFLAGS" CFLAGS="$CFLAGS" LDFLAGS="$LDFLAGS" \
               LIBS="$mikmod_dependencies $flac_dependencies"

  # check for MSYS permission errors
  if [ x"`grep 'Permission denied' config.log`" != x ]; then
      echo '**** MSYS problems; build aborted.'
      exit 1
  fi

  # ./configure messes up on its Makefile generation, putting some rules
  # on the same line as their targets, and placing multiple targets in one
  # line. Break them up. Also add the required FLAC and mikmod linkage flags.
  mv -f Makefile Makefile~
  sed -e 's~\\(-c $< -o $@\\) \\($(objects)\\)~\\1\\\n\\2~g' \
      -e 's~\\(\\.c\\)\\(\t$(LIBTOOL)\\)~\\1\\\n\\2~g' \
      -e 's~\\(: \\./version.rc\\)\\(\t$(WINDRES)\\)~\\1\\\n\\2~' \
      -e "s~\\(-lFLAC\\)~\\1 $flac_dependencies~" \
      -e "s~\\(-lmikmod\\)~\\1 $mikmod_dependencies~" \
      Makefile~ >Makefile

  cd ..
fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  case $BDTARGET in
    x86) windres="windres --target=pe-i386";;
    x64) windres="windres --target=pe-x86-64";;
    *) echo Unknown target format $BDTARGET; exit 1;;
  esac
  # Only build the DLL. The executables link to WinMain, which is not found.
  make build/libSDL_mixer.la WINDRES="$windres"
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install-hdrs install-lib
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [[ x$BDCLEAN == x1 && -f Makefile ]]; then
  rm -rf $BDSUBDIR
fi
"""),
    Dependency('PORTMIDI', ['portmidi', 'portmidi-[1-9].*'], ['portmidi.dll'], """

set -e
export PATH="$PREFIX/bin:$PATH"

if [ x$BDCONF == x1 ]; then
  rm -rf $BDSUBDIR
  mkdir $BDSUBDIR
  cp -r -t $BDSUBDIR pm_common porttime pm_win

  # Fix up some g++ 4.5.0 issues in the source code.
  source_file=pm_common/portmidi.c
  sed \
-e '420,+7s/return !Pm_QueueEmpty(midi->queue);/\
return Pm_QueueEmpty(midi->queue) ? pmNoData : pmGotData;/' \
      $source_file >$BDSUBDIR/$source_file

  source_file=pm_common/pmutil.c
  sed \
-e '45s/bzero(\\([^,]\\+,\\)/memset(\\1 0,/' \
      $source_file >$BDSUBDIR/$source_file

  source_file=pm_win/pmwin.c
  sed \
-e '20,+6s/^\\(#include <windows.h>\\)/#include <ctype.h>\\\n\\1/' \
-e '91,+7s/if (RegQueryValueEx(hkey, key, NULL, &dwType, pattern, &pattern_max) !=/\
if (RegQueryValueEx(hkey, key, NULL, \\&dwType, (BYTE *) pattern, (DWORD *) \\&pattern_max) !=/' \
      $source_file >$BDSUBDIR/$source_file

  source_file=pm_win/pmwinmm.c
  sed \
-e '40,+9s/DWORD/DWORD_PTR/g' \
-e '164s/UINT/UINT_PTR/' \
-e '209s/UINT/UINT_PTR/' \
-e '212s/ = / = (MIDIOUTCAPS \\*) /' \
-e '536s/\\(return\\)\\( \\+[^;]\\+\\);/\\1 pm_hosterror ?\\2 : pmNoError;/' \
-e '551s/DWORD/DWORD_PTR/' \
-e '670,+2s/DWORD/DWORD_PTR/g' \
-e '831,+2s/DWORD/DWORD_PTR/g' \
-e '895s/max/__max/' \
-e '1211s/ = / = (uint32_t \\*)/' \
-e '1301,+1s/DWORD/DWORD_PTR/g' \
-e '1331s/DWORD/DWORD_PTR/g' \
-e '1427s/ = / = (PmInternal \\*)/' \
      $source_file >$BDSUBDIR/$source_file

  cat > $BDSUBDIR/GNUmakefile << THE_END
# Makefile for portmidi, generated for Pygame by $BDPROG
# on $(date).

PREFIX = $PREFIX

pmcom = pm_common
pmwin = pm_win
pt = porttime

pmdll = portmidi.dll
pmlib = libportmidi.a
pmimplib = libportmidi.dll.a
pmcomsrc = \\$(pmcom)/portmidi.c \\$(pmcom)/pmutil.c
pmwinsrc = \\$(pmwin)/pmwin.c \\$(pmwin)/pmwinmm.c
pmobj = portmidi.o pmutil.o pmwin.o pmwinmm.o
pmsrc = \\$(pmcomsrc) \\$(pmwinsrc)
pmreqhdr = \\$(pmcom)/portmidi.h \\$(pmcom)/pmutil.h
pmhdr = \\$(pmreqhdr) \\$(pmcom)/pminternal.h \\$(pmwin)/pmwinmm.h

ptsrc = \\$(pt)/porttime.c porttime/ptwinmm.c
ptobj = porttime.o ptwinmm.o
ptreqhdr = \\$(pt)/porttime.h
pthdr = \\$(ptreqhdr)

src = \\$(pmsrc) \\$(ptsrc)
reqhdr = \\$(pmreqhdr) \\$(ptreqhdr)
hdr = \\$(pmhdr) \\$(pthdr)
obj = \\$(pmobj) \\$(ptobj)
def = portmidi.def

IHDR := -I\\$(pmcom) -I\\$(pmwin) -I\\$(pt)
LIBS := \\$(LOADLIBES) \\$(LDLIBS) -lwinmm



all : \\$(pmdll)
.PHONY : all

\\$(pmlib) : \\$(src) \\$(hdr)
\tg++ $CPPFLAGS \\$(IHDR) -c $CFLAGS -fpermissive \\$(src)
\tar rc \\$(pmlib) \\$(obj)
\tranlib \\$(pmlib)

\\$(pmdll) : \\$(pmlib) \\$(def)
\tg++ -shared $LDFLAGS -def \\$(def) \\$(pmlib) \\$(LIBS) -o \\$@
\tdlltool -D \\$(pmdll) -d \\$(def) -l \\$(pmimplib)
\tranlib \\$(pmimplib)

.PHONY : install

install : \\$(pmdll)
\tcp -f --target-directory=\\$(PREFIX)/bin $<
\tcp -f --target-directory=\\$(PREFIX)/lib \\$(pmlib)
\tcp -f --target-directory=\\$(PREFIX)/lib \\$(pmimplib)
\tcp -f --target-directory=\\$(PREFIX)/include \\$(reqhdr)

.PHONY : clean

clean :
\trm -f \\$(obj) \\$(pmdll) \\$(pmimplib) \\$(pmlib)
THE_END

  cat > $BDSUBDIR/portmidi.def << 'THE_END'
LIBRARY portmidi.dll
EXPORTS
Pm_Abort
Pm_Close
Pm_CountDevices
Pm_Dequeue
Pm_Enqueue
Pm_GetDefaultInputDeviceID
Pm_GetDefaultOutputDeviceID
Pm_GetDeviceInfo
Pm_GetErrorText
Pm_GetHostErrorText
Pm_HasHostError
Pm_Initialize
Pm_OpenInput
Pm_OpenOutput
Pm_Poll
Pm_QueueCreate
Pm_QueueDestroy
Pm_QueueEmpty
Pm_QueueFull
Pm_QueuePeek
Pm_Read
Pm_SetChannelMask
Pm_SetFilter
Pm_SetOverflow
Pm_Terminate
Pm_Write
Pm_WriteShort
Pm_WriteSysEx
Pt_Sleep
Pt_Start
Pt_Started
Pt_Stop
Pt_Time
THE_END

fi

if [ x$BDCOMP == x1 ]; then
  cd $BDSUBDIR
  make
  cd ..
fi

if [ x$BDINST == x1 ]; then
  cd $BDSUBDIR
  make install
  cd ..
fi

if [ x$BDSTRIP == x1 ]; then
  for nextdll in $BDDLLS ; do
    strip --strip-all "$PREFIX/bin/$nextdll"
  done
fi

if [[ x$BDCLEAN == x1 && -f Makefile ]]; then
  rm -rf $BDSUBDIR
fi
"""),

    ]  # End dependencies = [.


msys_prep = Preparation('destintation directory', """

# Ensure the destination directory and its subdirectories exist.
mkdir -p "$PREFIX/lib"
mkdir -p "$PREFIX/include"
mkdir -p "$PREFIX/bin"
mkdir -p "$PREFIX/doc"
mkdir -p "$PREFIX/man"
mkdir -p "$PREFIX/share"

echo
echo \\$ ls -Fp1 "\\"$PREFIX\\""
ls -Fp1 "$PREFIX"
""")

report_environment = Preparation('build environment', """

# list environment variables
echo
echo "PATH                               : $PATH"
echo "PREFIX (target directory)          : $PREFIX"
echo "BDTARGET (target machine type)     : $BDTARGET"
echo "BDBUILD (GNU configure --build)    : $BDBUILD"
echo "BDHOST (GNU configure --host)      : $BDHOST"
echo "CPPFLAGS                           : $CPPFLAGS"
echo "CFLAGS                             : $CFLAGS"
echo "LDFLAGS                            : $LDFLAGS"
echo "INCLUDE                            : $INCLUDE"
echo "BDCONF (Configure?)                : $BDCONF"
echo "BDCOMP (Compile?)                  : $BDCOMP"
echo "BDINST (Install?)                  : $BDINST"
echo "BDSTRIP (Strip?)                   : $BDSTRIP"
echo "BDCLEAN (Clean?)                   : $BDCLEAN"
echo "BDRESOURCES (resource object file) : $BDRESOURCES"
echo "BDNBEEPS (number of alert beeps)   : $BDNBEEPS"
echo "BDSUBDIR (build subdirectory)      : $BDSUBDIR"
""")

alert = Preparation('computer beeper', r"""

#
# Alert the user by beeping the computer speaker
#
for (( i = $BDNBEEPS ; i ; i-- )); do
    printf $'\a\a\a\a\a'
    sleep 1s
done
""")

if __name__ == '__main__':
    sys.exit(main(dependencies, msys_prep))
