from msysio import print_, raw_input_

from distutils.extension import read_setup_file
from optparse import OptionParser
import os
import re
import sys

# For Python 2.x/3.x compatibility
def geterror():
    return sys.exc_info()[1]

parent_dir_wp = os.path.dirname(os.getcwd())

prebuilt_dir = 'prebuilt'
lib_subdir = 'lib'

def get_ptr_size():
    return 64 if sys.maxsize > 2**32 else 32

def as_machine_type(size):
    """Return pointer bit size as a Windows machine type"""
    if size == 32:
        return "x86"
    if size == 64:
        return "x64"
    raise BuildError("Unknown pointer size {}".format(size))

class MakePrebuiltError(Exception):
    pass

def file_copy(src, dest):
    if os.path.isdir(dest):
        dest = os.path.join(dest, os.path.split(src)[1])
    with open(src, 'rb') as s:
        with open(dest, 'wb') as d:
            d.write(s.read())
            print_("%s => %s" % (src, dest))

def find_import_libraries(path, roots):
    match = re.compile(r'lib(%s)\.dll\.a$' % '|'.join(roots)).match
    return [name for name in os.listdir(path) if match(name) is not None]

def copy_dir(src, dest):
    if dest == '.':
        ignore, dest = os.path.split(src)
    if src == dest:
        raise IOError("%s: Source and destination are identical" % src)
    makedirs(dest)
    for name in os.listdir(src):
        src_path = os.path.join(src, name)
        if os.path.isfile(src_path):
            file_copy(src_path, os.path.join(dest, name))

def confirm(message):
    "ask a yes/no question, return result"
    reply = raw_input_("\n%s [Y/n]:" % message)
    if reply and reply[0].lower() == 'n':
        return 0
    return 1

created_dirs = set()

def makedirs(path):
    path = os.path.abspath(path)
    if path in created_dirs:
        pass
    elif not os.path.exists(path):
        os.makedirs(path)
        created_dirs.add(path)
    elif not os.path.isdir(path):
        raise MakePrebuiltError("%s is not a directory" % path)


def command_line():
    """Process the command line and return the options"""

    usage = ("usage: %prog [options] [destination]\n"
             "\n"
             "Assemble the Python 2.5 prebuilt dependencies directory. The\n"
             "default destination is .\\prebuilt .\n"
             "\n"
             "At startup this program may prompt for missing information.\n"
             "Be aware of this before redirecting output or leaving the\n"
             "program unattended. Once the '=== Assembling ===' message appears\n"
             "no more user input is required.")

    parser = OptionParser(usage)
    parser.add_option('-f', '--force', action='store_true', dest='force',
                      help="Overwrite an existing destination directory"
                           " without prompting first")
    parser.set_defaults(force=False)
    parser.add_option('-s', '--source', action='store',
                      dest='source_directory',
                      help="Directory where the DLLs and headers are installed")
    parser.set_defaults(source_directory='')
    parser.add_option('-t', '--target', type='choice', choices=['x86', 'x64'],
                      dest='target_type',
                      help="Build for x86 or x64 Windows")
    parser.set_defaults(target_type=as_machine_type(get_ptr_size()))
    return parser.parse_args()

def main(dest_dir=None, src_dir=None, force=None, target=None):
    # Top level directories.
    if dest_dir is None:
        dest_dir = ''
    if src_dir is None:
        source_dir = ''
    if force is None:
        force = False
    if target is None:
        target = as_machine_type(get_ptr_size());
    if not dest_dir:
        dest_dir = os.path.join(parent_dir_wp, 'prebuilt-' + target)
    if re.match(r'([A-Za-z]:){0,1}[^"<>:|?*]+$', dest_dir) is None:
        print_("Invalid directory path name %s" % (dest_dir,))
        return 1
    dest_dir = os.path.abspath(dest_dir)
    if not force and os.path.isdir(dest_dir):
        if not confirm("Directory %s already exists;\ncontinue" % dest_dir):
            return 1
    makedirs(dest_dir)
    if not src_dir:
        src_dir = os.path.join(parent_dir_wp, 'build-' + target)
    else:
        src_dir = os.path.abspath(src_dir)
    if not os.path.isdir(src_dir):
        print_("Source directory %s not found." % (src_dir,))
        return 1
    print_("=== Assembling ===")
    prebuilt_template = os.path.abspath('templates')
    dest_lib_dir = os.path.join(dest_dir, lib_subdir)

    # Read setup file.
    src_file = os.path.join(prebuilt_template, 'Setup_Win.in')
    dst_file = os.path.join(dest_dir, 'Setup_Win.in')
    with open(src_file, 'r') as setup_win:
        setup_win_tmpl = setup_win.read()
    with open(dst_file, 'w') as setup_win:
        prebuilt_dir = 'prebuilt-' + target
        setup_win.write(setup_win_tmpl.format(prebuilt=prebuilt_dir))
    deps =  read_setup_file(dst_file)
    with open(dst_file) as setup_win:
        match = re.compile('[A-Z_][A-Z0-9_]* *=(.*)').match
        header_dir_pat = re.compile(' -I([^ ]+)')
        lib_pat = re.compile(' -l([^ ]+)')
        macros = []
        for line in setup_win:
            matches = match(line)
            if matches is not None:
                flags = matches.group(1)
                header_dirs = header_dir_pat.findall(flags)
                libs = lib_pat.findall(flags)
                macros.append((header_dirs, libs))

    # Copy DLLs.
    src_bin_dir = os.path.join(src_dir, 'bin')
    have_dlls = set()
    for dep in deps:
        path_elements = dep.library_dirs[0].split('/')  # / required by setup.
        dll_name = path_elements[-1]
        src_dll_path = os.path.join(src_bin_dir, dll_name)
        if os.path.exists(src_dll_path):
            if path_elements[0] == '.':
                path_elements = path_elements[2:]
            else:
                path_elements = path_elements[1:]
            dest_dll_dir = dest_dir
            for dir_name in path_elements[:-1]:
                dest_dll_dir = os.path.join(dest_dll_dir, dir_name)
                makedirs(dest_dll_dir)
            file_copy(os.path.join(src_bin_dir, dll_name),
                      os.path.join(dest_dll_dir, dll_name))
            have_dlls.add(dep.name[8:])

    # Copy required import libraries only.
    copied_files = set()
    src_lib_dir = os.path.join(src_dir, 'lib')
    makedirs(dest_lib_dir)
    for ignore, libs in macros:
        use = False
        for lib in libs:
            if lib in have_dlls:
                use = True
                break
        if use and lib not in copied_files:
            copied_files.add(lib)
            lib_name = 'lib%s.dll.a' % lib
            src_lib_path = os.path.join(src_lib_dir, lib_name)
            if not os.path.exists(src_lib_path):
                print_("Missing import library %s" % (lib_name,))
                return 1
            file_copy(src_lib_path, os.path.join(dest_lib_dir, lib_name))

    # Copy required header directories only.
    copied_dirs = set()
    for header_dirs, libs in macros:
        use = False
        for lib in libs:
            if lib in have_dlls:
                use = True
                break
        if use:
            for header_dir in header_dirs:
                path_elements = header_dir.split('/')
                if path_elements[0] == '.':
                    path_elements = path_elements[2:]
                else:
                    path_elements = path_elements[1:]
                src_header_dir = os.path.join(src_dir, *path_elements)
                if not os.path.exists(src_header_dir):
                    print_("Missing include directory %s" % (src_header_dir,))
                    return 1
                dest_header_dir = dest_dir
                for dir_name in path_elements:
                    dest_header_dir = os.path.join(dest_header_dir, dir_name)
                    makedirs(dest_header_dir)
                if not src_header_dir in copied_dirs:
                    copy_dir(src_header_dir, dest_header_dir)
                    copied_dirs.add(src_header_dir)
    if 'SDL' in have_dlls:
        # For MSVC use SDL_config_win32.h in place of configure
        # generated SDL_config.h.
        file_copy(
            os.path.join(src_dir, 'include', 'SDL', 'SDL_config_win32.h'),
            os.path.join(dest_dir, 'include', 'SDL', 'SDL_config.h'))

    vcvars = 'vcvars{}.bat'.format('64' if target == 'x64' else '32')    

    # Def file bat.
    with open(os.path.join(dest_lib_dir, 'MakeDefs.bat'), 'w') as make_defs:
        make_defs.write(
            '@echo off\n'
            'rem Make .def files needed for .lib file creation.\n'
            'rem Requires gendef.exe on the search path\n'
            'rem (found under mingw-w64-tools in the MinGW-w64\n'
            'rem git repository at SourceForge:\n'
            'rem http://sourceforge.net/p/mingw-w64/mingw-w64/ci/master/tree/'
            ').\n\n')
        for dep in deps:
            dll_name = os.path.split(dep.library_dirs[0])[1]
            lib = dep.name[8:]
            lib_name = 'lib%s.dll.a' % lib
            if os.path.exists(os.path.join(dest_lib_dir, lib_name)):
                start = ''
            else:
                start = 'rem '
            make_defs.write('%sgendef - %s >%s.def\n' %
                            (start, dll_name, lib))

    # Lib import files bat.
    with open(os.path.join(dest_lib_dir, 'MakeLibs.bat'), 'w') as make_libs:
        hdr = ('@echo off\n'
               'rem Make .lib import libraries.\n'
               'rem Requires Visual C++ Studio or Toolkit.\n'
               'rem {} must be run first to use LINK.EXE.\n\n'.format(vcvars))
        make_libs.write(hdr)
        for dep in deps:
            dll_name = os.path.split(dep.library_dirs[0])[1]
            lib = dep.name[8:]
            lib_name = 'lib%s.dll.a' % lib
            if os.path.exists(os.path.join(dest_lib_dir, lib_name)):
                start = ''
            else:
                start = 'rem '
            make_libs.write('%sLINK.EXE /LIB /NOLOGO /DEF:%s.def /MACHINE:%s /OUT:%s.lib\n' %
                            (start, lib, target, lib))

    # Top level make batch file.
    with open(os.path.join(dest_lib_dir, 'Make.bat'), 'w') as make:
        content = ('@echo off\n'
                   'rem Builds the Visual C {machine} import libraries.\n'
                   'rem Requires gendef (see MakeDefs.bat) and {vcvars} on\n'
                   'rem the executable search path.\n'
                   '\n'
                   'CALL {vcvars}\n'
                   'CALL MakeDefs.bat\n'
                   'CALL MakeLibs.bat\n'.format(machine=target, vcvars=vcvars))
        make.write(content)

    return 0

if __name__ =='__main__':
    options, args = command_line()
    dest_dir = None
    if args:
        dest_dir = args[0]
    try:
        sys.exit(main(dest_dir=dest_dir,
                      src_dir=options.source_directory,
                      force=options.force,
                      target=options.target_type))
    except MakePrebuiltError:
        print_("*** %s; execution halted" % (geterror(),))
        sys.exit(1)
